/*
wav2raw - convert RIFF/RIFX WAVE PCM to Plan 9s "raw" file format (See audio(3))
Copyright (C) 2005 Sascha Retzki.
released under the terms of the MIT-License.


*/

#include <u.h>
#include <libc.h>

#include <libaudioconv.h>


int outputsamplerate = 0;


int debuglevel = 0;

void
debug(char *msg)
{
	if(debuglevel > 0)
		fprint(2,msg);
}



/*
We do the work here. We need to know all the infos the FMT-chunk gave us, plus the sample-rate that we are supposed
to convert to. As we know the application logic, we can read from stdin ourselves. Note that process() is split up - the
actual processing could have been done in one loop. But it is not really intelligent to check if the input is 16bit or not on
each iteration. (Tho I am pretty sure that testing if the input is stereo everytime is also pretty stupid.
*/

void
process(long samplerate, ACriffwave_fmt *fmt)
{
	long byterate = samplerate * 2 * 2, rr;
	long safedatasize = (fmt->byterate > byterate ? fmt->byterate : byterate) * sizeof(short);
	short *datacopy;



	if((datacopy = malloc(safedatasize)) == 0)
		sysfatal("%r");


	if(fmt->bitspersample == 16) {
		short *data;

		if((data = malloc(safedatasize)) == 0)
			sysfatal("%r");

		while((rr = readn(0,data,fmt->byterate * sizeof(short)-1)) != 0) {


			if(fmt->channels == 1)
				convertmono2stereo(data, datacopy, rr/sizeof(short));

			if(fmt->samplerate != samplerate)
				convertsamplerate(data, datacopy, fmt->samplerate, samplerate,rr/sizeof(short));

			if(rr == fmt->byterate) {
				if(write(1,data,byterate) != byterate)
					sysfatal("%r");
			} else {
				long outsize = ((float) rr / (float)fmt->byterate) * byterate;
				if(write(1,data,outsize) != outsize)
					sysfatal("%r");
			}
		}
		free(data);
	} else if(fmt->bitspersample == 8) {
		char *indata; // TODO: unsigned char ...?
		short *outdata;

		if((indata = malloc(fmt->byterate)) == 0)
			sysfatal("%r");

		if((outdata = malloc(safedatasize)) == 0)
			sysfatal("%r");

		while((rr = readn(0,indata,fmt->byterate)) != 0) {


			convert8bitspersample(indata, outdata, fmt->byterate);

			if(fmt->channels == 1)
				convertmono2stereo(outdata, datacopy, rr/sizeof(short));

			if(fmt->samplerate != samplerate)
				convertsamplerate(outdata, datacopy, fmt->samplerate, samplerate,rr/sizeof(short));

			if(rr == fmt->byterate) {
				if(write(1,outdata,byterate) != byterate)
					sysfatal("%r");
			} else {
				long outsize = ((float) rr / (float)fmt->byterate) * byterate;
				if(write(1,outdata,outsize) != outsize)
					sysfatal("%r");
			}
		}
		free(indata);
		free(outdata);
	} else {
		sysfatal("process(): Sorry. Currently, %d bits per sample are not supported\n",fmt->bitspersample);
	}


	free(datacopy);
}




void
usage(void)
{
	print("usage: %s [-d] [-p] [-s sample_rate]\n",argv0);


	print("-d\t\tincrease debug level\n");

	print("-s long\tconvert to this sample rate (default 44100)\n");

	print("-p\t\tjust parse FMT-header (implies and thus increments the debug level)\n");

	print("data is expected on stdin and passed to stdout\n");
	print("metadata (-d, etc) is written to stderr\n");

	exits("usage");
}

void
main(int argc, char **argv)
{
	/* data needs to be unsigned because of the FMT-header. signed char is not big enough
		e.g. to hold a sample-rate of 44100hz. */
	unsigned char *data, datahead[]= "data";
	int datasearch = 0, justparse = 0, byteorder;
	ACriffwave_fmt *fmt = malloc(sizeof(ACriffwave_fmt));

	ARGBEGIN {
		case 'd':
			debuglevel++;
			break;
		case 's':
			if((outputsamplerate = atoi(EARGF(usage()))) == 0)
				sysfatal("ASCII to Integer-convertion of argument to -s failed");
			break;
		case 'p':
			justparse = 1;
			debuglevel++;
			break;
		default:
			usage();
	} ARGEND
	


	/* RIFF-WAVE header... */
	if((data = malloc(12)) == 0)
		sysfatal("%r");
	if(readn(0,data,12) != 12)
		sysfatal("%r (EOF?)");
	if((byteorder = parseriffwave(data)) == ACError)
		sysfatal("no RIFF/RIFX-WAVE header...");


	/* Get the FMT-chunk */
	if(readn(0,data,4) != 4)
		sysfatal("%r while searching for FMT-chunk (EOF?)");
	if(strncmp((char *)data,"fmt ",4) == 0) {
		debug("found \"fmt \", investigating...\n");
		if((data = realloc(data,20)) == nil)
			sysfatal("%r");
		if(readn(0,data,20) != 20)
			sysfatal("%r while reading FMT-chunk");
		if(!parsefmt(data, byteorder,fmt))
			sysfatal("FMT-chunk contains unexpected/idiotic values");
	} else
		sysfatal("could not find the FMT-chunk at the expected offset (found %s)\n",(char *)data);


	/* Print something informational. This looks ugly but you ain't debug it, so just skip it :P */
	debug(smprint("FMT-Header...:\nAudioformat %d\nChannels %d\n",fmt->audioformat, fmt->channels));
	debug(smprint("Samplerate %d\nByterate %d\n", fmt->samplerate, fmt->byterate));
	debug(smprint("Bockalign %d\nBitspersample %d\n",fmt->blockalign,fmt->bitspersample));


	/* do we actually have more work to do? */
	if(justparse)
		exits(nil);


	/* now we directly search for the data-chunk (we just support PCM, anyway..)
	     funfact: Wintendos sound recorder puts a FACT-chunk right after FMT even if it is PCM, which
	     is at least idiotic; believing various resources, it is even non-standard. Welcome to Hell... */
	while(1) {
		if(readn(0,data,1) != 1)
			sysfatal("%r while searching for DATA-chunk (EOF?)");
		if(datahead[datasearch] == data[0])
			datasearch++;
		if(datasearch == 4) { 
			if(readn(0,data,4) != 4)
				sysfatal("%r while reading datachunk/chunksize (EOF? No data actually?)");
			break;
		}
	}
	free(data);


	process(outputsamplerate > 0 ? outputsamplerate : 44100, fmt);


	exits(nil);
}