/*
riffxwave - RIFF/RIFX WAVE parser
Copyright (C) 2005 Sascha Retzki.
released under the terms of the MIT-License.

Credits:
- Wikipedia for giving a brief info and linking me to:
- http://ccrma.stanford.edu/CCRMA/Courses/422/projects/WaveFormat/
- libsndfile for both, a bit of info and creating test-files
- Microsoft(R)(Tm)(C)(w00t) for creating this file form... or rather for distributing 
	the Sound recorder which created the other half of test-files for this program.
- Audacity, THE Lunix/Wintendo-world DAE
- and last but never least Russ Cox for teaching me C

Notes:
- The code has not been tested with RIFX. I did not yet find a RIFX-Wave file. 
- I am sure there are PCM-like codecs out there which would work - but just PCM is strictly supported by the FMT-parser.
*/

#include <u.h>
#include <libc.h>

#include "libaudioconv.h"





int
parseriffwave(unsigned char *data)
{
	int returnvalue;


	if(!strncmp((char *)data,"RIFF",4)) {
		returnvalue = ACRIFF;
		goto wave;
	} else if(!strncmp((char *)data,"RIFX",4)) {
		returnvalue = ACRIFX;
		goto wave;
	} else
		return ACError;

wave:
	data += 8; 
	if(!strncmp((char *)data,"WAVE",4))
		return returnvalue;
	else 
		return ACError;
}

int
parsefmt(unsigned char *data, int byteorder, ACriffwave_fmt *fmt)
{

	if(byteorder == ACRIFF) {
		fmt->audioformat = data[5] << 8 | data[4];
		fmt->channels = data[7] << 8 | data[6];
		fmt->samplerate = data[11] << 24 | data[10] << 16 | data[9] << 8 | data[8];
		fmt->byterate = data[15] << 24 | data[14] << 16 | data[13] << 8 | data[12];
		fmt->blockalign = data[17] << 8 | data[16];
		fmt->bitspersample = data[19] << 8 | data[18];
	} else if(byteorder == ACRIFX) {
		fmt->audioformat = data[4] << 8 | data[5];
		fmt->channels = data[6] << 8 | data[7];
		fmt->samplerate = data[8] << 24 | data[9] << 16 | data[10] << 8 | data[11];
		fmt->byterate = data[12] << 24 | data[13] << 16 | data[14] << 8 | data[15];
		fmt->blockalign = data[16] << 8 | data[17];
		fmt->bitspersample = data[18] << 8 | data[19];
	} else
		sysfatal("Hit the programmer with a cluestick, please. Suplied byte-order value unknown.");



	// Is this too strict? One could say we check if we did everything wrong, but this could also be a task for the app

	if(fmt->audioformat != 1) {
		return ACEaudioformat;
	} else if(fmt->channels <= 0) {
		return ACEchannels;
	} else if(fmt->samplerate <= 0) {
		return ACEsamplerate;
	} else if(fmt->bitspersample%8 != 0) {
		return ACEbitspersample;
	} else if(fmt->byterate != (fmt->samplerate * fmt->channels * (fmt->bitspersample/8))) {
		return ACEbyterate;
	} else
		return 1;
}



char *
mkriffxwave(int headertype, int filesize)
{
	char *data = malloc(ACHSRIFFWAVE);
	

	data[0] = 'R';
	data[1] = 'I';
	data[2] = 'F';
	switch(headertype) {
		case ACRIFF:
			data[3] = 'F';
			break;
		case ACRIFX:
			data[3] = 'X';
			break;
		default:
			sysfatal("libaudioconv/mkriffwave(): unknown headertype");
	}


	data[4] = (char) filesize;
	data[5] = (char) (filesize >> 8);
	data[6] = (char) (filesize >> 16);
	data[7] = (char ) (filesize >> 24);

	data[8] = 'W';
	data[9] = 'A';
	data[10]='V';
	data[11]='E'; 

	return data;			
}

unsigned char *
mkfmt(int headertype, ACriffwave_fmt *fmt)
{
	int sizeoffmt = 16;
	unsigned char *data = malloc(ACHSFMT);

	data[0] = 'f';
	data[1] = 'm';
	data[2] = 't';
	data[3] = ' ';

	/* Size is little endian on RIFX, too? */
	data[4] = (char) sizeoffmt;
	data[5] = (char) (sizeoffmt >> 8);
	data[6] = (char) (sizeoffmt >> 16);
	data[7] = (char ) (sizeoffmt >> 24);


	if(headertype == ACRIFF) {


		data[8] = (char) fmt->audioformat;
		data[9] = (char) (fmt->audioformat >> 8);

		data[10] = (char) fmt->channels;
		data[11] = (char) (fmt->channels >> 8);

		data[12] = (char) fmt->samplerate;
		data[13] = (char) (fmt->samplerate >> 8);
		data[14] = (char) (fmt->samplerate >> 16);
		data[15] = (char ) (fmt->samplerate >> 24);

		data[16] = (char) fmt->byterate;
		data[17] = (char) (fmt->byterate >> 8);
		data[18] = (char) (fmt->byterate >> 16);
		data[19] = (char ) (fmt->byterate >> 24);

		data[20] = (char) fmt->blockalign;
		data[21] = (char) (fmt->blockalign >> 8);

		data[22] = (char) fmt->bitspersample;
		data[23] = (char) (fmt->bitspersample >> 8);


	} else if(headertype == ACRIFX) {

		data[8] = (char) (fmt->audioformat >> 8);
		data[9] = (char) fmt->audioformat;


		data[10] = (char) (fmt->channels >> 8);
		data[11] = (char) fmt->channels;

		data[12] = (char) (fmt->samplerate >> 24);
		data[13] = (char) (fmt->samplerate >> 16);
		data[14] = (char) (fmt->samplerate >> 8);
		data[15] = (char ) fmt->samplerate;


		data[16] = (char) (fmt->byterate >> 24);
		data[17] = (char) (fmt->byterate >> 16);
		data[18] = (char) (fmt->byterate >> 8);
		data[19] = (char ) fmt->byterate;

		data[20] = (char) (fmt->blockalign >> 8);
		data[21] = (char) fmt->blockalign;

		data[22] = (char) (fmt->bitspersample >> 8);
		data[23] = (char) fmt->bitspersample;

	} else {
		sysfatal("libaudioconv/mkfmt(): Unknown headertype: %d",headertype);
	}

	return data;
}

char *
mkdata(int filesize)
{
	char *data = malloc(ACHSDATA);

	data[0] = 'D';
	data[1] = 'A';
	data[2] = 'T';
	data[3] = 'A';


	// DATA-filesize == filesize - sizeof(RIFF-WAVE chunk) - sizeof(FMT-chunk)
	data[4] = (char) (filesize -36);
	data[5] = (char) ((filesize -36) >> 8);
	data[6] = (char) ((filesize -36) >> 16);
	data[7] = (char ) ((filesize -36) >> 24);


	return data;
}